class SpiDev(object):
    def __init__(self, bus, device):
        self.bus = bus
        self.device = device
        self.max_speed_hz = 0
        self.mode = 0

    def writebytes(self, data):
        pass



try:
    import picamera
except ImportError:
    import fake_rpi.picamera as picamera

import io
import logging
import threading
import time

from .hub import Hub
from .hub import HubMessage
from .hub import HubMessageType
from .keypad import KeyTypes


class Capture(threading.Thread):
    def __init__(self, hub: Hub) -> None:
        super().__init__()
        self.daemon = True
        self.camera: picamera.PiCamera
        self._stop_evt = threading.Event()
        self._data_queue = hub.bus
        hub.add_subscriber(self)
        self._message = HubMessage(HubMessageType.DATA, bytes())
        self._zoom_factor = 0
        self._events = {
            KeyTypes.KEY1: lambda: None,  # Turn on capture
            KeyTypes.KEY2: lambda: None,  # Mirror image
            KeyTypes.KEY3: lambda: None,  # TBD
            KeyTypes.UP: lambda: self._zoom(0.1),    # camera zoom in
            KeyTypes.DOWN: lambda: self._zoom(-0.1),  # camera zoom out
            KeyTypes.LEFT: lambda: self._rotate(-90),  # Rotate left
            KeyTypes.RIGHT: lambda: self._rotate(90),  # Rotate right
            KeyTypes.PRESS: lambda: self._reset()  # Reset settings
        }

    def run(self) -> None:
        self.camera = picamera.PiCamera()
        self.camera.resolution = (1920, 1080)
        self.camera.framerate = 30
        time.sleep(2)  # camera initialization
        stream = io.BytesIO()
        for _ in self.camera.capture_continuous(stream, 'jpeg', use_video_port=True):
            try:
                stream.seek(0)
                self._message.content = stream.read()
                self._data_queue.put_nowait(self._message)
            except Exception:
                pass
            finally:
                stream.seek(0)
                stream.truncate()
            if self._stop_evt.is_set():
                break
        self.camera.close()
        self._stop_evt.clear()

    def stop(self) -> None:
        self._stop_evt.set()

    def on_message(self, message: HubMessage):
        if message.type == HubMessageType.EVENT:
            event_type = message.content
            logging.debug('Event received: %r', event_type)
            self._events.get(event_type, lambda: None)()

    def _rotate(self, inc):
        self.camera.rotation = self.camera.rotation + inc

    def _zoom(self, inc):
        zf = self._zoom_factor
        zf += inc
        if zf > 0.4:
            zf = 0.4
        if zf < 0:
            zf = 0
        if zf == self._zoom_factor:
            return
        self._zoom_factor = zf
        logging.debug('Setting zoom factor to: %f', zf)
        self.camera.zoom = (0+zf, 0+zf, 1-2*zf, 1-2*zf)

    def _reset(self):
        self._zoom_factor = 0
        self.camera.zoom = (0, 0, 1, 1)
        self.camera.rotation = 0



import enum
import threading

try:
    import RPi.GPIO as GPIO
except ImportError:
    from fake_rpi.RPi import GPIO


from .hub import Hub
from .hub import HubMessage
from .hub import HubMessageType


class KeyTypes(enum.IntEnum):
    LEFT = 5
    UP = 6
    PRESS = 13
    DOWN = 19
    RIGHT = 26
    KEY1 = 21
    KEY2 = 20
    KEY3 = 16


class KeyPad(threading.Thread):
    def __init__(self, hub: Hub) -> None:
        super().__init__()
        self._stop_evt = threading.Event()
        self._message_bus = hub.bus
        self._message = HubMessage(HubMessageType.EVENT, None)

    def run(self) -> None:
        self._gpio_setup()
        while not self._stop_evt.is_set():
            self._stop_evt.wait(30)
        self._stop_evt.clear()
        self._cleanup()

    def stop(self) -> None:
        self._stop_evt.set()

    def _gpio_setup(self):
        GPIO.setmode(GPIO.BCM)
        for key in KeyTypes:
            GPIO.setup(key, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            GPIO.add_event_detect(key, GPIO.FALLING, callback=self._on_key_pressed, bouncetime=200)

    def _on_key_pressed(self, key):
        key_type = KeyTypes(key)
        self._message.content = key_type
        self._message_bus.put_nowait(self._message)

    def _cleanup(self):
        for key in KeyTypes:
            GPIO.remove_event_detect(key)



import io
import logging
import queue
import threading

from PIL import Image

from . import lcd
from .hub import Hub
from .hub import HubMessage
from .hub import HubMessageType

LCD_SIZE = lcd.LCD_WIDTH, lcd.LCD_HEIGHT


class Display(threading.Thread):
    def __init__(self, hub: Hub) -> None:
        super().__init__()
        self._data_queue = queue.Queue()
        self._stop_evt = threading.Event()
        self._lcd = lcd.LCD()
        self._lcd.LCD_Init(lcd.SCAN_DIR_DFT)
        self._lcd.LCD_Clear()
        self._lcd.LCD_Off()
        self._background = Image.new('RGBA', LCD_SIZE, (0, 0, 0, 0))
        self._displaying = False
        hub.add_subscriber(self)

    def on_message(self, message: HubMessage):
        if self._displaying:
            logging.debug('Previous image is still being rendered, skipping data')
            return
        if message.type == HubMessageType.DATA:
            self._data_queue.put(message.content)

    def run(self):
        self._lcd.LCD_On()
        while not self._stop_evt.is_set():
            try:
                data = self._data_queue.get(30)
                self._displaying = True
                self._display(data)
            except queue.Empty:
                pass
            finally:
                self._displaying = False
        self._stop_evt.clear()

    def stop(self):
        self._stop_evt.set()

    def _display(self, data: bytes):
        image = Image.open(io.BytesIO(data))
        image.thumbnail(LCD_SIZE)
        position = (int((LCD_SIZE[0] - image.size[0]) / 2), int((LCD_SIZE[1] - image.size[1]) / 2))
        self._background.paste(image, position)
        self._lcd.LCD_ShowImage(self._background, 0, 0)
        logging.debug('Image Rendered')



import enum
import queue
import threading
from abc import ABCMeta
from typing import Any


class HubSubscriber(metaclass=ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return hasattr(subclass, 'on_message') and callable(subclass.on_message) or NotImplemented


class HubMessageType(enum.Enum):
    DATA = 1
    EVENT = 2


class HubMessage():
    def __init__(self, msg_type: HubMessageType, content: Any) -> None:
        super().__init__()
        self.type: HubMessageType = msg_type
        self.content: Any = content


class Hub(threading.Thread):
    def __init__(self) -> None:
        super().__init__()
        self._subscribers = set()
        self._bus = queue.Queue()
        self._stop_evt = threading.Event()

    def add_subscriber(self, subscriber: HubSubscriber):
        if not isinstance(subscriber, HubSubscriber):
            raise ValueError('Bad interface')
        self._subscribers.add(subscriber)

    def run(self) -> None:
        while not self._stop_evt.is_set():
            try:
                data = self._bus.get(5)
                self._distribute(data)
            except queue.Empty:
                pass
        self._stop_evt.clear()

    def stop(self) -> None:
        self._stop_evt.set()
        self.join()

    def _distribute(self, message: HubMessage) -> None:
        subscriber: HubSubscriber
        for subscriber in self._subscribers:
            subscriber.on_message(message)

    @property
    def bus(self):
        return self._bus



##
#  @filename   :   DEV_Config.py
#  @brief      :   LCD hardware interface implements (GPIO, SPI)
#  @author     :   Yehui from Waveshare
#
#  Copyright (C) Waveshare     July 10 2017
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documnetation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to  whom the Software is
# furished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS OR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

try:
    import spidev
except ImportError:
    from . import fake_spi as spidev

try:
    import RPi.GPIO as GPIO
except ImportError:
    from fake_rpi.RPi import GPIO

import time

# Pin definition
LCD_RST_PIN = 27
LCD_DC_PIN = 25
LCD_CS_PIN = 8
LCD_BL_PIN = 24

# SPI device, bus = 0, device = 0
SPI = spidev.SpiDev(0, 0)


def epd_digital_write(pin, value):
    GPIO.output(pin, value)


def Driver_Delay_ms(xms):
    time.sleep(xms / 1000.0)


def SPI_Write_Byte(data):
    SPI.writebytes(data)


def GPIO_Init():
    GPIO.setmode(GPIO.BCM)
    GPIO.setwarnings(False)
    GPIO.setup(LCD_RST_PIN, GPIO.OUT)
    GPIO.setup(LCD_DC_PIN, GPIO.OUT)
    GPIO.setup(LCD_CS_PIN, GPIO.OUT)
    GPIO.setup(LCD_BL_PIN, GPIO.OUT)
    SPI.max_speed_hz = 9000000
    SPI.mode = 0b00
    return 0



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mspihole` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``spihole.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``spihole.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import functools
import logging
import os
import signal
import sys
import threading

import click

from .capture import Capture
from .display import Display
from .hub import Hub
from .keypad import KeyPad


def signal_handler(stop_event: threading.Event, received_signal, _frame):
    logging.info('Signal received: %s', received_signal)
    stop_event.set()


def install_handlers(stop_event: threading.Event):
    handler = functools.partial(signal_handler, stop_event)
    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGTERM, handler)


@click.command()
@click.option("-c", "--configuration", type=click.Path(readable=True),
              default=os.path.join(os.path.sep, 'etc', 'spihole.conf'))
def main(configuration):
    # configuration_file = click.format_filename(configuration)
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG,
                        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    hub = Hub()
    capture = Capture(hub)
    display = Display(hub)
    keypad = KeyPad(hub)

    startables = [hub, display, capture, keypad]
    for startable in startables:
        startable.start()

    stop_evt = threading.Event()
    logging.debug('Starting main loop')
    while not stop_evt.is_set():
        try:
            signal.pause()  # interruptible idle
        except KeyboardInterrupt:
            stop_evt.set()
    logging.debug('Main loop completed, closing threads')
    for startable in startables:
        startable.stop()

    for startable in startables:
        startable.join()



# -*- coding:UTF-8 -*-
##
# | file      	:	LCD_1IN44.py
# |	version		:	V2.0
# | date		:	2018-07-16
# | function	:	On the ST7735S chip driver and clear screen, drawing lines, drawing, writing
# 					and other functions to achieve
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documnetation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to  whom the Software is
# furished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS OR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
try:
    import RPi.GPIO as GPIO
except ImportError:
    from fake_rpi.RPi import GPIO

import numpy as np

from . import config as LCD_Config

LCD_1IN44 = 1
LCD_1IN8 = 0
if LCD_1IN44 == 1:
    LCD_WIDTH = 128  # LCD width
    LCD_HEIGHT = 128  # LCD height
    LCD_X = 2
    LCD_Y = 1
if LCD_1IN8 == 1:
    LCD_WIDTH = 160
    LCD_HEIGHT = 128
    LCD_X = 1
    LCD_Y = 2

LCD_X_MAXPIXEL = 132  # LCD width maximum memory
LCD_Y_MAXPIXEL = 162  # LCD height maximum memory

# scanning method
L2R_U2D = 1
L2R_D2U = 2
R2L_U2D = 3
R2L_D2U = 4
U2D_L2R = 5
U2D_R2L = 6
D2U_L2R = 7
D2U_R2L = 8
SCAN_DIR_DFT = U2D_R2L


class LCD:
    def __init__(self):
        self.width = LCD_WIDTH
        self.height = LCD_HEIGHT
        self.LCD_Scan_Dir = SCAN_DIR_DFT
        self.LCD_X_Adjust = LCD_X
        self.LCD_Y_Adjust = LCD_Y

    """    Hardware reset     """

    def LCD_Reset(self):
        GPIO.output(LCD_Config.LCD_RST_PIN, GPIO.HIGH)
        LCD_Config.Driver_Delay_ms(100)
        GPIO.output(LCD_Config.LCD_RST_PIN, GPIO.LOW)
        LCD_Config.Driver_Delay_ms(100)
        GPIO.output(LCD_Config.LCD_RST_PIN, GPIO.HIGH)
        LCD_Config.Driver_Delay_ms(100)

    """    Write register address and data     """

    def LCD_WriteReg(self, Reg):
        GPIO.output(LCD_Config.LCD_DC_PIN, GPIO.LOW)
        LCD_Config.SPI_Write_Byte([Reg])

    def LCD_WriteData_8bit(self, Data):
        GPIO.output(LCD_Config.LCD_DC_PIN, GPIO.HIGH)
        LCD_Config.SPI_Write_Byte([Data])

    def LCD_WriteData_NLen16Bit(self, Data, DataLen):
        GPIO.output(LCD_Config.LCD_DC_PIN, GPIO.HIGH)
        for i in range(0, DataLen):
            LCD_Config.SPI_Write_Byte([Data >> 8])
            LCD_Config.SPI_Write_Byte([Data & 0xFF])

    """    Common register initialization    """

    def LCD_InitReg(self):
        # ST7735R Frame Rate
        self.LCD_WriteReg(0xB1)
        self.LCD_WriteData_8bit(0x01)
        self.LCD_WriteData_8bit(0x2C)
        self.LCD_WriteData_8bit(0x2D)

        self.LCD_WriteReg(0xB2)
        self.LCD_WriteData_8bit(0x01)
        self.LCD_WriteData_8bit(0x2C)
        self.LCD_WriteData_8bit(0x2D)

        self.LCD_WriteReg(0xB3)
        self.LCD_WriteData_8bit(0x01)
        self.LCD_WriteData_8bit(0x2C)
        self.LCD_WriteData_8bit(0x2D)
        self.LCD_WriteData_8bit(0x01)
        self.LCD_WriteData_8bit(0x2C)
        self.LCD_WriteData_8bit(0x2D)

        # Column inversion
        self.LCD_WriteReg(0xB4)
        self.LCD_WriteData_8bit(0x07)

        # ST7735R Power Sequence
        self.LCD_WriteReg(0xC0)
        self.LCD_WriteData_8bit(0xA2)
        self.LCD_WriteData_8bit(0x02)
        self.LCD_WriteData_8bit(0x84)
        self.LCD_WriteReg(0xC1)
        self.LCD_WriteData_8bit(0xC5)

        self.LCD_WriteReg(0xC2)
        self.LCD_WriteData_8bit(0x0A)
        self.LCD_WriteData_8bit(0x00)

        self.LCD_WriteReg(0xC3)
        self.LCD_WriteData_8bit(0x8A)
        self.LCD_WriteData_8bit(0x2A)
        self.LCD_WriteReg(0xC4)
        self.LCD_WriteData_8bit(0x8A)
        self.LCD_WriteData_8bit(0xEE)

        self.LCD_WriteReg(0xC5)  # VCOM
        self.LCD_WriteData_8bit(0x0E)

        # ST7735R Gamma Sequence
        self.LCD_WriteReg(0xE0)
        self.LCD_WriteData_8bit(0x0F)
        self.LCD_WriteData_8bit(0x1A)
        self.LCD_WriteData_8bit(0x0F)
        self.LCD_WriteData_8bit(0x18)
        self.LCD_WriteData_8bit(0x2F)
        self.LCD_WriteData_8bit(0x28)
        self.LCD_WriteData_8bit(0x20)
        self.LCD_WriteData_8bit(0x22)
        self.LCD_WriteData_8bit(0x1F)
        self.LCD_WriteData_8bit(0x1B)
        self.LCD_WriteData_8bit(0x23)
        self.LCD_WriteData_8bit(0x37)
        self.LCD_WriteData_8bit(0x00)
        self.LCD_WriteData_8bit(0x07)
        self.LCD_WriteData_8bit(0x02)
        self.LCD_WriteData_8bit(0x10)

        self.LCD_WriteReg(0xE1)
        self.LCD_WriteData_8bit(0x0F)
        self.LCD_WriteData_8bit(0x1B)
        self.LCD_WriteData_8bit(0x0F)
        self.LCD_WriteData_8bit(0x17)
        self.LCD_WriteData_8bit(0x33)
        self.LCD_WriteData_8bit(0x2C)
        self.LCD_WriteData_8bit(0x29)
        self.LCD_WriteData_8bit(0x2E)
        self.LCD_WriteData_8bit(0x30)
        self.LCD_WriteData_8bit(0x30)
        self.LCD_WriteData_8bit(0x39)
        self.LCD_WriteData_8bit(0x3F)
        self.LCD_WriteData_8bit(0x00)
        self.LCD_WriteData_8bit(0x07)
        self.LCD_WriteData_8bit(0x03)
        self.LCD_WriteData_8bit(0x10)

        # Enable test command
        self.LCD_WriteReg(0xF0)
        self.LCD_WriteData_8bit(0x01)

        # Disable ram power save mode
        self.LCD_WriteReg(0xF6)
        self.LCD_WriteData_8bit(0x00)

        # 65k mode
        self.LCD_WriteReg(0x3A)
        self.LCD_WriteData_8bit(0x05)

    # ********************************************************************************
    # function:	Set the display scan and color transfer modes
    # parameter:
    # 		Scan_dir   :   Scan direction
    # 		Colorchose :   RGB or GBR color format
    # ********************************************************************************
    def LCD_SetGramScanWay(self, Scan_dir):
        # Get the screen scan direction
        self.LCD_Scan_Dir = Scan_dir

        # Get GRAM and LCD width and height
        if (
            (Scan_dir == L2R_U2D)
            or (Scan_dir == L2R_D2U)
            or (Scan_dir == R2L_U2D)
            or (Scan_dir == R2L_D2U)
        ):
            self.width = LCD_HEIGHT
            self.height = LCD_WIDTH
            if Scan_dir == L2R_U2D:
                MemoryAccessReg_Data = 0x00 | 0x00
            elif Scan_dir == L2R_D2U:
                MemoryAccessReg_Data = 0x00 | 0x80
            elif Scan_dir == R2L_U2D:
                MemoryAccessReg_Data = 0x40 | 0x00
            else:  # R2L_D2U:
                MemoryAccessReg_Data = 0x40 | 0x80
        else:
            self.width = LCD_WIDTH
            self.height = LCD_HEIGHT
            if Scan_dir == U2D_L2R:
                MemoryAccessReg_Data = 0x00 | 0x00 | 0x20
            elif Scan_dir == U2D_R2L:
                MemoryAccessReg_Data = 0x00 | 0x40 | 0x20
            elif Scan_dir == D2U_L2R:
                MemoryAccessReg_Data = 0x80 | 0x00 | 0x20
            else:  # R2L_D2U
                MemoryAccessReg_Data = 0x40 | 0x80 | 0x20

        # please set (MemoryAccessReg_Data & 0x10) != 1
        if (MemoryAccessReg_Data & 0x10) != 1:
            self.LCD_X_Adjust = LCD_Y
            self.LCD_Y_Adjust = LCD_X
        else:
            self.LCD_X_Adjust = LCD_X
            self.LCD_Y_Adjust = LCD_Y

        # Set the read / write scan direction of the frame memory
        self.LCD_WriteReg(0x36)  # MX, MY, RGB mode
        if LCD_1IN44 == 1:
            self.LCD_WriteData_8bit(MemoryAccessReg_Data | 0x08)  # 0x08 set RGB
        else:
            self.LCD_WriteData_8bit(
                MemoryAccessReg_Data & 0xF7
            )  # RGB color filter panel

    # /********************************************************************************
    # function:
    # 			initialization
    # ********************************************************************************/
    def LCD_Init(self, Lcd_ScanDir):
        if LCD_Config.GPIO_Init() != 0:
            return -1

        # Turn on the backlight
        GPIO.output(LCD_Config.LCD_BL_PIN, GPIO.HIGH)

        # Hardware reset
        self.LCD_Reset()

        # Set the initialization register
        self.LCD_InitReg()

        # Set the display scan and color transfer modes
        self.LCD_SetGramScanWay(Lcd_ScanDir)
        LCD_Config.Driver_Delay_ms(200)

        # sleep out
        self.LCD_WriteReg(0x11)
        LCD_Config.Driver_Delay_ms(120)

        # Turn on the LCD display
        self.LCD_WriteReg(0x29)

    def LCD_Off(self):
        GPIO.output(LCD_Config.LCD_BL_PIN, GPIO.LOW)

    def LCD_On(self):
        GPIO.output(LCD_Config.LCD_BL_PIN, GPIO.HIGH)

    # /********************************************************************************
    # function:	Sets the start position and size of the display area
    # parameter:
    # 	Xstart 	:   X direction Start coordinates
    # 	Ystart  :   Y direction Start coordinates
    # 	Xend    :   X direction end coordinates
    # 	Yend    :   Y direction end coordinates
    # ********************************************************************************/
    def LCD_SetWindows(self, Xstart, Ystart, Xend, Yend):
        # set the X coordinates
        self.LCD_WriteReg(0x2A)
        self.LCD_WriteData_8bit(0x00)
        self.LCD_WriteData_8bit((Xstart & 0xFF) + self.LCD_X_Adjust)
        self.LCD_WriteData_8bit(0x00)
        self.LCD_WriteData_8bit(((Xend - 1) & 0xFF) + self.LCD_X_Adjust)

        # set the Y coordinates
        self.LCD_WriteReg(0x2B)
        self.LCD_WriteData_8bit(0x00)
        self.LCD_WriteData_8bit((Ystart & 0xFF) + self.LCD_Y_Adjust)
        self.LCD_WriteData_8bit(0x00)
        self.LCD_WriteData_8bit(((Yend - 1) & 0xFF) + self.LCD_Y_Adjust)

        self.LCD_WriteReg(0x2C)

    def LCD_Clear(self):
        # hello
        _buffer = [0xFF] * (self.width * self.height * 2)
        self.LCD_SetWindows(0, 0, self.width, self.height)
        GPIO.output(LCD_Config.LCD_DC_PIN, GPIO.HIGH)
        for i in range(0, len(_buffer), 4096):
            LCD_Config.SPI_Write_Byte(_buffer[i:i + 4096])

    def LCD_ShowImage(self, Image, Xstart, Ystart):
        if Image is None:
            return
        imwidth, imheight = Image.size
        if imwidth != self.width or imheight != self.height:
            raise ValueError(
                "Image must be same dimensions as display \
                ({0}x{1}).".format(
                    self.width, self.height
                )
            )
        img = np.asarray(Image)
        pix = np.zeros((self.width, self.height, 2), dtype=np.uint8)
        pix[..., [0]] = np.add(
            np.bitwise_and(img[..., [0]], 0xF8), np.right_shift(img[..., [1]], 5)
        )
        pix[..., [1]] = np.add(
            np.bitwise_and(np.left_shift(img[..., [1]], 3), 0xE0),
            np.right_shift(img[..., [2]], 3),
        )
        pix = pix.flatten().tolist()
        self.LCD_SetWindows(0, 0, self.width, self.height)
        GPIO.output(LCD_Config.LCD_DC_PIN, GPIO.HIGH)
        for i in range(0, len(pix), 4096):
            LCD_Config.SPI_Write_Byte(pix[i:i + 4096])
